﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace HaloMap
{
    public class CollisionMesh
    {
        #region Fields
        public List<Vector3> Vertices = new List<Vector3>();
        public List<ushort> Faces = new List<ushort>();
        public Reflexive BSP = new Reflexive();
        public ScriptReference Piece = new ScriptReference();
        public ScriptReference Permutation = new ScriptReference();
        #endregion
    }

    public class PathFindingSphere
    {
        #region Fields
        public int NodeIndex;
        public Vector3 Position = new Vector3();
        public float Radius;
        #endregion
    }

    public class Collision
    {
        #region Fields
        public List<IndexBuffer> IndexBuffers = new List<IndexBuffer>();
        public List<VertexBuffer> VertexBuffers = new List<VertexBuffer>();
        public List<Mesh> SphereMesh = new List<Mesh>();
        public List<Matrix> SphereMatrix = new List<Matrix>();
        public List<int> VerticeCount = new List<int>();
        public List<int> FaceCount = new List<int>();

        public List<CollisionMesh> Meshes = new List<CollisionMesh>();
        public List<PathFindingSphere> Spheres = new List<PathFindingSphere>();
        public ScriptReference Piece = new ScriptReference();
        public ScriptReference Permutation = new ScriptReference();
        #endregion

        #region Constructor
        public Collision(Map map)
        {
            //Clear all fields
            IndexBuffers.Clear();
            VertexBuffers.Clear();
            SphereMesh.Clear();
            SphereMatrix.Clear();
            VerticeCount.Clear();
            FaceCount.Clear();
            Meshes.Clear();
            Spheres.Clear();

            //Regions
            Reflexive Regions = new Reflexive();
            Regions.Read(map, 28);
            for (int i = 0; i < Regions.ChunkCount; i++)
            {
                //Permutations
                map.br.BaseStream.Position = Regions.Translation + (i * 12);
                Piece.Read(map, true);
                Reflexive Permutations = new Reflexive();
                Permutations.Read(map, true);
                for (int x = 0; x < Permutations.ChunkCount; x++)
                {
                    //BSPs
                    map.br.BaseStream.Position = Permutations.Translation + (x * 20);
                    Permutation.Read(map, true);
                    Reflexive BSP = new Reflexive();
                    BSP.Read(map, true);
                    for (int z = 0; z < BSP.ChunkCount; z++)
                    {
                        //Faces
                        CollisionMesh Mesh = new CollisionMesh();
                        Mesh.BSP = BSP;
                        Mesh.Permutation = Permutation;
                        Mesh.Piece = Piece;
                        map.br.BaseStream.Position = BSP.Translation + (z * 68) + 52;
                        Reflexive f = new Reflexive();
                        f.Read(map, true);
                        for (int y = 0; y < f.ChunkCount; y++)
                        {
                            map.br.BaseStream.Position = f.Translation + (y * 12);
                            Mesh.Faces.Add(map.br.ReadUInt16());
                            Mesh.Faces.Add(map.br.ReadUInt16());
                            Mesh.Faces.Add(Mesh.Faces[(y * 3) + 1]);
                        }

                        //Verts
                        map.br.BaseStream.Position = BSP.Translation + (z * 68) + 60;
                        f.Read(map, true);
                        for (int y = 0; y < f.ChunkCount; y++)
                        {
                            map.br.BaseStream.Position = f.Translation + (y * 16);
                            Vector3 V3 = new Vector3();
                            V3.X = map.br.ReadSingle();
                            V3.Y = map.br.ReadSingle();
                            V3.Z = map.br.ReadSingle();
                            Mesh.Vertices.Add(V3);
                        }
                        Meshes.Add(Mesh);
                    }
                }
            }

            //Pathfinding Spheres
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 36;
            Reflexive Sphere = new Reflexive();
            Sphere.Read(map, true);
            for (int i = 0; i < Sphere.ChunkCount; i++)
            {
                map.br.BaseStream.Position = Sphere.Translation + (20 * i);
                PathFindingSphere S = new PathFindingSphere();
                S.NodeIndex = map.br.ReadInt32();
                S.Position.X = map.br.ReadSingle();
                S.Position.Y = map.br.ReadSingle();
                S.Position.Z = map.br.ReadSingle();
                S.Radius = map.br.ReadSingle();
                Spheres.Add(S);
            }
        }
        #endregion

        #region Methods
        public void Intialize(Map map, ref Device Device)
        {
            //Spheres
            for (int i = 0; i < Spheres.Count; i++)
            {
                Mesh m = Mesh.Sphere(Device, Spheres[i].Radius, 10, 10);
                SphereMesh.Add(m);
                Matrix matrix = Matrix.Identity;
                matrix.Multiply(Matrix.Translation(Spheres[i].Position));
                SphereMatrix.Add(matrix);
            }

            //Index Buffers
            for (int i = 0; i < Meshes.Count; i++)
            {
                if (Meshes[i].Permutation == Meshes[0].Permutation)
                {
                    IndexBuffer buffer = new IndexBuffer(typeof(short), Meshes[i].Faces.Count, Device, Usage.WriteOnly, Pool.Default);
                    buffer.SetData(Meshes[i].Faces, 0, LockFlags.None);
                    buffer.Unlock();
                    IndexBuffers.Add(buffer);
                    FaceCount.Add(Meshes[i].Faces.Count);
                }
            }

            //Vertex Buffers
            for (int i = 0; i < Meshes.Count; i++)
            {
                if (Meshes[i].Permutation == Meshes[0].Permutation)
                {
                    VertexBuffer vbx = new VertexBuffer(typeof(CustomVertex.PositionColored), Meshes[i].Vertices.Count, Device, Usage.WriteOnly, CustomVertex.PositionColored.Format, Pool.Default);
                    CustomVertex.PositionColored[] verts = (CustomVertex.PositionColored[])vbx.Lock(0, 0);
                    for (int x = 0; x < Meshes[i].Vertices.Count; x++)
                    {
                        verts[x].Position = Meshes[i].Vertices[x];
                    }
                    vbx.Unlock();
                    VertexBuffers.Add(vbx);
                    VerticeCount.Add(Meshes[i].Vertices.Count);
                }
            }
        }

        public void DrawSpheres(ref Device Device)
        {
            for (int i = 0; i < SphereMesh.Count; i++)
            {
                Device.RenderState.AlphaBlendEnable = false;
                Device.RenderState.AlphaTestEnable = false;
                Device.Transform.World = SphereMatrix[i];
                Device.SetTexture(0, null);
                if (!SphereMesh[i].Disposed)
                {
                    SphereMesh[i].DrawSubset(0);
                }
            }
        }

        public void DrawMeshes(ref Device Device)
        {
            for (int i = 0; i < VertexBuffers.Count; i++)
            {
                Device.SetStreamSource(0, VertexBuffers[i], 0);
                Device.VertexFormat = CustomVertex.PositionColored.Format;
                Device.Indices = IndexBuffers[i];
                Device.RenderState.AlphaBlendEnable = false;
                Device.RenderState.AlphaTestEnable = false;
                Device.Transform.World = Matrix.Identity;
                Device.SetTexture(0, null);
                Device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, VerticeCount[i], 0, FaceCount[i] / 3);
            }
        }

        public void Extract(Map map, string Location)
        {
            StreamWriter sw = new StreamWriter(new FileStream(Location, FileMode.Create, FileAccess.Write, FileShare.Write));
            sw.WriteLine("#---------------------------------------------------------");
            sw.WriteLine("# Collision Mesh - Extracted with Gravemind");
            sw.WriteLine("#---------------------------------------------------------");

            for (int i = 0; i < Meshes.Count; i++)
            {
                //Verts
                for (int x = 0; x < Meshes[i].Vertices.Count; x++)
                {
                    Vector3 v = Meshes[i].Vertices[x];
                    sw.WriteLine("v " + v.X.ToString() + " " + v.Y.ToString() + " " + v.Z.ToString());
                }

                //Faces
                for (int x = 0; x < Meshes[i].Faces.Count; x += 3)
                {
                    sw.WriteLine("f " + (Meshes[i].Faces[x] + 1).ToString()
                        + " " + (Meshes[i].Faces[x + 1] + 1).ToString()
                        + " " + (Meshes[i].Faces[x + 2] + 1).ToString());
                }
            }

            sw.Close();
        }

        public void Dispose()
        {
            IndexBuffers = null;
            VertexBuffers = null;
            SphereMesh = null;
            SphereMatrix = null;
            Spheres = null;
            VerticeCount = null;
            FaceCount = null;
            Meshes = null;
            Piece = null;
            Permutation = null;
            GC.Collect();
        }
        #endregion
    }
}
