﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace HaloMap
{
    public enum DDSEnum : int
    {
        DDSD_CAPS = 0x1,
        DDSD_HEIGHT = 0x2,
        DDSD_WIDTH = 0x4,
        DDSD_PITCH = 0x8,
        DDSD_PIXELFORMAT = 0x1000,
        DDSD_MIPMAPCOUNT = 0x20000,
        DDSD_LINEARSIZE = 0x80000,
        DDSD_DEPTH = 0x800000,

        DDPF_ALPHAPIXELS = 0x1,
        DDPF_FOURCC = 0x4,
        DDPF_RGB = 0x40,

        //'The dwCaps1 member of the DDSCAPS2 structure can be set to one or more of the following values.
        DDSCAPS_COMPLEX = 0x8,
        DDSCAPS_TEXTURE = 0x1000,
        DDSCAPS_MIPMAP = 0x400000,

        //The dwCaps2 member of the DDSCAPS2 structure can be set to one or more of the following values.
        DDSCAPS2_CUBEMAP = 0x200,
        DDSCAPS2_CUBEMAP_POSITIVEX = 0x400,
        DDSCAPS2_CUBEMAP_NEGATIVEX = 0x800,
        DDSCAPS2_CUBEMAP_POSITIVEY = 0x1000,
        DDSCAPS2_CUBEMAP_NEGATIVEY = 0x2000,
        DDSCAPS2_CUBEMAP_POSITIVEZ = 0x4000,
        DDSCAPS2_CUBEMAP_NEGATIVEZ = 0x8000,
        DDSCAPS2_VOLUME = 0x200000
    }

    public struct DDS_Header
    {
        #region Fields
        public string Magic;
        public DDS_SurfaceDesc dds_SurfaceDesc;
        #endregion

        #region Methods
        public void ReadStruct(ref BinaryReader br)
        {
            Magic = new string(br.ReadChars(4));
            dds_SurfaceDesc.ReadStruct(ref br);
        }

        public void WriteStruct(ref BinaryWriter br)
        {
            br.Write(Magic.ToCharArray());
            dds_SurfaceDesc.WriteStruct(ref br);
        }

        public void Generate(ref BitmapChunk bi)
        {
            Magic = "DDS ";
            dds_SurfaceDesc.Generate(ref bi);
        }
        #endregion
    }

    public struct DDS_SurfaceDesc
    {
        #region Fields
        public int size_of_structure;
        public int flags;
        public int height;
        public int width;
        public int PitchOrLinearSize;
        public int depth; //Only used for volume textures.
        public int MipMapCount; //Total Number of MipMaps
        public int[] Reserved1;
        public DDS_PixelFormat dds_PixelFormat;
        public DDS_Caps dds_Caps;
        public int Reserved2;
        #endregion

        #region Methods
        public void ReadStruct(ref BinaryReader br)
        {
            size_of_structure = br.ReadInt32();
            flags = br.ReadInt32();
            height = br.ReadInt32();
            width = br.ReadInt32();
            PitchOrLinearSize = br.ReadInt32();
            depth = br.ReadInt32();
            MipMapCount = br.ReadInt32();

            Reserved1 = new int[11]; //11 ints long
            for (int x = 0; x < 11; x++)
                Reserved1[x] = br.ReadInt32();

            dds_PixelFormat.ReadStruct(ref br);
            dds_Caps.ReadStruct(ref br);

            Reserved2 = br.ReadInt32();
        }

        public void WriteStruct(ref BinaryWriter bw)
        {
            bw.Write(size_of_structure);
            bw.Write(flags);
            bw.Write(height);
            bw.Write(width);
            bw.Write(PitchOrLinearSize);
            bw.Write(depth);
            bw.Write(MipMapCount);

            for (int x = 0; x < 11; x++)
                bw.Write(Reserved1[x]);

            dds_PixelFormat.WriteStruct(ref bw);
            dds_Caps.WriteStruct(ref bw);

            bw.Write(Reserved2);
        }

        // Generate the DDSURFACEDESC2 chunk of the header
        public void Generate(ref BitmapChunk bi)
        {
            Reserved1 = new int[11];

            //'Size of structure. This member must be set to 124.
            size_of_structure = 124;

            //Flags to indicate valid fields. Always include DDSD_CAPS, DDSD_PIXELFORMAT,
            //DDSD_WIDTH, DDSD_HEIGHT and either DDSD_PITCH or DDSD_LINEARSIZE.
            flags = flags + (int)DDSEnum.DDSD_CAPS;
            flags = flags + (int)DDSEnum.DDSD_PIXELFORMAT;
            flags = flags + (int)DDSEnum.DDSD_WIDTH;
            flags = flags + (int)DDSEnum.DDSD_HEIGHT;

            switch (bi.BitmapFormat)
            {
                case BitmapFormat.DXT1:
                case BitmapFormat.DXT3:
                case BitmapFormat.DXT5:
                    flags = flags + (int)DDSEnum.DDSD_LINEARSIZE;
                    break;

                default:
                    flags = flags + (int)DDSEnum.DDSD_PITCH;
                    break;
            }

            //Height of the main image in pixels
            height = bi.Height;

            //Width of the main image in pixels
            width = bi.Width;

            //For uncompressed formats, this is the number of bytes per scan line (DWORD aligned) for the main
            //image. dwFlags should include DDSD_PITCH in this case. For compressed formats, this is the (total)
            //number of bytes for the main image. dwFlags should be include DDSD_LINEARSIZE in this case.
            int RGBBitCount = 0;
            switch (bi.BitmapFormat)
            {
                case BitmapFormat.R5G6B5:
                case BitmapFormat.A1R5G5B5:
                case BitmapFormat.A4R4G4B4:
                    RGBBitCount = 16;
                    break;

                case BitmapFormat.P8:
                case BitmapFormat.X8R8G8B8:
                case BitmapFormat.A8R8G8B8:
                    RGBBitCount = 32;
                    break;
            }

            PitchOrLinearSize = bi.Width * (RGBBitCount / 8);

            //For volume textures, this is the depth of the volume.
            //dwFlags should include DDSD_DEPTH in this case.
            depth = 0; // There are no volume textures (that I know of) in Halo.

            //For items with mipmap levels, this is the total number of levels in the mipmap
            //chain of the main image. dwFlags should include DDSD_MIPMAPCOUNT in this case
            MipMapCount = bi.MipMapCount;
            if (MipMapCount > 0) flags = flags + (int)DDSEnum.DDSD_MIPMAPCOUNT;

            //A 32-byte value that specifies the pixel format structure.
            dds_PixelFormat.Generate(bi);

            //A 16-byte value that specifies the capabilities structure.
            dds_Caps.Generate(ref bi);
        }
        #endregion
    }

    public struct DDS_PixelFormat
    {
        #region Fields
        public int size; //32
        public int Flags; //4
        public string FourCC; //DXT1, DXT2, etc..
        public BitmapFormat Format;
        public int RGBBitCount;
        public int RBitMask;
        public int GBitMask;
        public int BBitMask;
        public int RGBAlphaBitMask;
        #endregion

        #region Methods
        public void ReadStruct(ref BinaryReader br)
        {
            size = br.ReadInt32(); //32
            Flags = br.ReadInt32(); //4
            FourCC = new string(br.ReadChars(4)); //DXT1, DXT2, etc..
            RGBBitCount = br.ReadInt32();
            RBitMask = br.ReadInt32();
            GBitMask = br.ReadInt32();
            BBitMask = br.ReadInt32();
            RGBAlphaBitMask = br.ReadInt32();
            //Get Teture Type
            if (RGBBitCount == 16)
            {
                if (RBitMask == 0xF800 && GBitMask == 0x7E0 && BBitMask == 0x1F)
                {
                    Format = BitmapFormat.R5G6B5;
                }
                else if (RBitMask == 0x7C00 && GBitMask == 0x3E0 && BBitMask == 0x1F)
                {
                    Format = BitmapFormat.A1R5G5B5;
                }
                else if (RBitMask == 0xF00 && GBitMask == 0xF0 && BBitMask == 0xF)
                {
                    Format = BitmapFormat.A4R4G4B4;
                }
            }
            else if (RGBBitCount == 32)
            {
                if (RBitMask == 0xFF0000 && GBitMask == 0xFF00 && BBitMask == 0xFF && RGBAlphaBitMask == 0x0)
                {
                    Format = BitmapFormat.X8R8G8B8;
                }
                else if (RBitMask == 0xFF0000 && GBitMask == 0xFF00 && BBitMask == 0xFF && RGBAlphaBitMask == int.Parse("FF000000", System.Globalization.NumberStyles.HexNumber))
                {
                    //Format = (BitmapFormat.P8 | BitmapFormat.A8R8G8B8);
                    Format = BitmapFormat.A8R8G8B8;
                }
            }

            if (Flags == 4)
            {
                switch (FourCC)
                {
                    case "DXT1":
                        {
                            Format = BitmapFormat.DXT1;
                            break;
                        }
                    case "DXT2":
                    case "DXT3":
                        {
                            Format = BitmapFormat.DXT3;
                            break;
                        }
                    case "DXT4":
                    case "DXT5":
                        {
                            Format = BitmapFormat.DXT5;
                            break;
                        }
                }
            }
        }

        public void WriteStruct(ref BinaryWriter bw)
        {
            bw.Write(size);
            bw.Write(Flags);
            bw.Write(FourCC.ToCharArray());

            bw.Write(RGBBitCount);
            bw.Write(RBitMask);
            bw.Write(GBitMask);
            bw.Write(BBitMask);
            bw.Write(RGBAlphaBitMask);
        }

        // Generate the DDPIXELFORMAT chunk of the header
        public void Generate(BitmapChunk bi)
        {
            //Size of structure. This member must be set to 32. 
            size = 32;

            //Flags to indicate valid fields. Uncompressed formats will usually use DDPF_RGB to indicate
            //an RGB format, while compressed formats will use DDPF_FOURCC with a four-character code. 
            //This is accomplished in the below structure.

            //This is the four-character code for compressed formats. dwFlags should include DDPF_FOURCC in
            //this case. For DXTn compression, this is set to "DXT1", "DXT2", "DXT3", "DXT4", or "DXT5". 
            Flags = 0;
            switch (bi.BitmapFormat)
            {
                case BitmapFormat.DXT1:
                    FourCC = "DXT1";
                    Flags = Flags + (int)DDSEnum.DDPF_FOURCC;
                    break;
                case BitmapFormat.DXT3:
                    FourCC = "DXT3";
                    Flags = Flags + (int)DDSEnum.DDPF_FOURCC;
                    break;
                case BitmapFormat.DXT5:
                    FourCC = "DXT4";
                    Flags = Flags + (int)DDSEnum.DDPF_FOURCC;
                    break;
                default:
                    FourCC = "\0\0\0\0";
                    Flags = Flags + (int)DDSEnum.DDPF_RGB;
                    break;
            }

            //For RGB formats, this is the total number of bits in the format. dwFlags should include DDPF_RGB
            //in this case. This value is usually 16, 24, or 32. For A8R8G8B8, this value would be 32. 
            switch (bi.BitmapFormat)
            {
                case BitmapFormat.R5G6B5:
                    RGBBitCount = 16;
                    break;

                case BitmapFormat.A1R5G5B5:
                case BitmapFormat.A4R4G4B4:
                    RGBBitCount = 16;
                    Flags = Flags + (int)DDSEnum.DDPF_ALPHAPIXELS;
                    break;

                case BitmapFormat.X8R8G8B8:
                    //Flags = Flags + (int)DDSEnum.DDPF_ALPHAPIXELS;
                    RGBBitCount = 32;
                    break;

                case BitmapFormat.P8:
                case BitmapFormat.A8R8G8B8:
                    Flags = Flags + (int)DDSEnum.DDPF_ALPHAPIXELS;
                    RGBBitCount = 32;
                    break;

                default:
                    RGBBitCount = 0;
                    break;
            }

            //For RGB formats, this contains the masks for the red, green, and blue channels. For A8R8G8B8, these
            //values would be 0x00ff0000, 0x0000ff00, and 0x000000ff respectively. 
            switch (bi.BitmapFormat)
            {
                case BitmapFormat.R5G6B5:
                    RBitMask = 0xF800;
                    GBitMask = 0x7E0;
                    BBitMask = 0x1F;
                    RGBAlphaBitMask = 0x0;
                    break;

                case BitmapFormat.A1R5G5B5:
                    RBitMask = 0x7C00;
                    GBitMask = 0x3E0;
                    BBitMask = 0x1F;
                    RGBAlphaBitMask = 0x8000;
                    break;

                case BitmapFormat.A4R4G4B4:
                    RBitMask = 0xF00;
                    GBitMask = 0xF0;
                    BBitMask = 0xF;
                    RGBAlphaBitMask = 0xF000;
                    break;

                case BitmapFormat.X8R8G8B8:
                    RBitMask = 0xFF0000;
                    GBitMask = 0xFF00;
                    BBitMask = 0xFF;
                    RGBAlphaBitMask = 0x0;
                    break;

                case BitmapFormat.P8:
                case BitmapFormat.A8R8G8B8:
                    RBitMask = 0xFF0000;
                    GBitMask = 0xFF00;
                    BBitMask = 0xFF;
                    RGBAlphaBitMask = int.Parse("FF000000", System.Globalization.NumberStyles.HexNumber);
                    break;

                default:
                    RBitMask = 0;
                    GBitMask = 0;
                    BBitMask = 0;
                    RGBAlphaBitMask = 0x0;
                    break;
            }

            //For RGB formats, this contains the mask for the alpha channel, if any. dwFlags should include 
            //DDPF_ALPHAPIXELS in this case. For A8R8G8B8, this value would be 0xff000000. 
            //^^ This was set in the previous block of code.
        }
        #endregion
    }

    public struct DDS_Caps
    {
        #region Fields
        public int caps1;
        public int caps2;
        public int[] Reserved;
        #endregion

        #region Methods
        public void ReadStruct(ref BinaryReader br)
        {
            caps1 = br.ReadInt32();
            caps2 = br.ReadInt32();
            Reserved = new int[2];
            Reserved[0] = br.ReadInt32();
            Reserved[1] = br.ReadInt32();
        }

        public void WriteStruct(ref BinaryWriter bw)
        {
            bw.Write(caps1);
            bw.Write(caps2);
            bw.Write(0);//Reserved[0]);
            bw.Write(0); //Reserved[1]);
        }

        // Generate the DDSCAPS2 chunk of the header
        public void Generate(ref BitmapChunk b2)
        {
            //DDS files should always include DDSCAPS_TEXTURE. If the file contains mipmaps, DDSCAPS_MIPMAP
            //should be set. For any DDS file with more than one main surface,such as a mipmaps, cubic
            //environment map, or volume texture, DDSCAPS_COMPLEX should also be set. 
            caps1 = caps1 + (int)DDSEnum.DDSCAPS_TEXTURE;
            if (b2.MipMapCount > 0) caps1 = caps1 + (int)DDSEnum.DDSCAPS_MIPMAP;
            caps1 = caps1 + (int)DDSEnum.DDSCAPS_COMPLEX;

            //For cubic environment maps, DDSCAPS2_CUBEMAP should be included as well as one or more faces of
            //the map (DDSCAPS2_CUBEMAP_POSITIVEX, DDSCAPS2_CUBEMAP_NEGATIVEX, DDSCAPS2_CUBEMAP_POSITIVEY,
            // DDSCAPS2_CUBEMAP_NEGATIVEY, DDSCAPS2_CUBEMAP_POSITIVEZ, DDSCAPS2_CUBEMAP_NEGATIVEZ). For volume
            //textures, DDSCAPS2_VOLUME should be included. 
            //******************************************************************************
            //This is where I stopped - this code needs to be added for cubemaps.
            //******************************************************************************
            caps2 = 0;
        }
        #endregion
    }

    public class DDS
    {
        #region Methods
        public static void Extract(Map map, BitmapChunk Chunk, string Path)
        {
            //Create the File
            BinaryWriter bw = new BinaryWriter(new FileStream(Path + ".dds", FileMode.Create, FileAccess.Write, FileShare.Write));

            //Decompress, and Decode
            byte[] Data = Bitmap.Decode(map, Chunk);

            //Write
            DDS_Header dds_Header = new DDS_Header();
            dds_Header.Generate(ref Chunk);
            dds_Header.WriteStruct(ref bw);
            bw.Write(Data);
            bw.Close();
        }

        public static void Inject(Map map, BitmapChunk Chunk, string Path)
        {
            BinaryReader br = new BinaryReader(new FileStream(Path, FileMode.Open, FileAccess.Read, FileShare.Read));

            //Read the Header
            DDS_Header dds_Header = new DDS_Header();
            dds_Header.ReadStruct(ref br);

            //Add new values
            Chunk.Height = (ushort)dds_Header.dds_SurfaceDesc.height;
            Chunk.Width = (ushort)dds_Header.dds_SurfaceDesc.width;
            Chunk.MipMapCount = (ushort)dds_Header.dds_SurfaceDesc.MipMapCount;
            Chunk.Flags = (ushort)dds_Header.dds_SurfaceDesc.flags;
            Chunk.Depth = (ushort)dds_Header.dds_SurfaceDesc.depth;
            Chunk.BitmapFormat = dds_Header.dds_SurfaceDesc.dds_PixelFormat.Format;

            //Compress and Encode
            int OldLength = Chunk.SourceData.Length;
            br.BaseStream.Position = 128;
            Chunk.SourceData = br.ReadBytes((int)(br.BaseStream.Length - 128));
            Chunk.SourceData = Bitmap.Encode(map, Chunk);
            br.Close();

            //Inject
            byte[] EOF;
            int Shift = 0;
            int Padding = 0;
            if (OldLength >= Chunk.SourceData.Length)
            {
                if (Chunk.RawLocation == RawLocation.Internal)
                {
                    map.bw.BaseStream.Position = Chunk.LOD1;
                    map.bw.Write(Chunk.SourceData);
                    byte[] Space = new byte[OldLength - Chunk.SourceData.Length];
                    map.bw.Write(Space);
                }
                else
                {
                    map.br.BaseStream.Position = map.Header.IndexOffset;
                    EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                    map.bw.BaseStream.Position = map.Header.IndexOffset;
                    Chunk.LOD1 = (int)map.bw.BaseStream.Position;
                    map.bw.Write(Chunk.SourceData);
                    Padding = map.Padding();
                    map.bw.Write(new byte[Padding]);
                    map.bw.Write(EOF);
                    Shift = Chunk.SourceData.Length + Padding;
                    map.Header.IndexOffset += Shift;
                    map.bw.BaseStream.Position = 8;
                    map.bw.Write((int)map.bw.BaseStream.Length);
                    map.bw.BaseStream.Position += 4;
                    map.Index.Read(map);
                    map.TagHandler.Read(map);
                    map.SelectedTag = map.FindTag(map.SelectedTag.Identifier);
                }
            }
            else
            {
                if (Chunk.RawLocation == RawLocation.Internal)
                {
                    map.br.BaseStream.Position = Chunk.LOD1 + Chunk.Size1;
                    EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                    map.bw.BaseStream.Position = Chunk.LOD1;
                    map.bw.Write(Chunk.SourceData);
                    Padding = map.Padding();
                    map.bw.Write(new byte[Padding]);
                    map.bw.Write(EOF);
                    Shift = Chunk.SourceData.Length - OldLength;
                    int Index = map.FindIndex(map.SelectedTag.Identifier);
                    for (int i = Index + 1; i < map.Tags.Count; i++)
                    {
                        if (map.Tags[i].Class == "bitm")
                        {
                            Tag t = map.Tags[i];
                            Reflexive BitmapData = new Reflexive();
                            BitmapData.Read(map, t.MetaOffset + 68 + Shift, true);
                            for (int c = 0; c < BitmapData.ChunkCount; c++)
                            {
                                map.br.BaseStream.Position = BitmapData.Translation + (c * 116) + 28 + Shift;
                                int Offset1 = map.br.ReadInt32();
                                int Offset2 = map.br.ReadInt32();
                                int Offset3 = map.br.ReadInt32();
                                if (Offset1 > 0 && RawData.RawLocator(Offset1) == RawLocation.Internal)
                                {
                                    Offset1 += Shift;
                                    map.bw.BaseStream.Position = BitmapData.Translation + (c * 116) + 28 + Shift;
                                    map.bw.Write(Offset1);
                                }
                                if (Offset2 > 0 && RawData.RawLocator(Offset2) == RawLocation.Internal)
                                {
                                    Offset2 += Shift;
                                    map.bw.BaseStream.Position = BitmapData.Translation + (c * 116) + 28 + Shift + 4;
                                    map.bw.Write(Offset2);
                                }
                                if (Offset3 > 0 && RawData.RawLocator(Offset3) == RawLocation.Internal)
                                {
                                    Offset3 += Shift;
                                    map.bw.BaseStream.Position = BitmapData.Translation + (c * 116) + 28 + Shift + 8;
                                    map.bw.Write(Offset3);
                                }
                            }
                        }
                    }
                }
                else
                {
                    map.br.BaseStream.Position = map.Header.IndexOffset;
                    EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                    map.bw.BaseStream.Position = map.Header.IndexOffset;
                    Chunk.LOD1 = (int)map.bw.BaseStream.Position;
                    map.bw.Write(Chunk.SourceData);
                    Padding = map.Padding();
                    map.bw.Write(new byte[Padding]);
                    map.bw.Write(EOF);
                    Shift = Chunk.SourceData.Length + Padding;
                    map.Header.IndexOffset += Shift;
                    map.bw.BaseStream.Position = 8;
                    map.bw.Write((int)map.bw.BaseStream.Length);
                    map.bw.BaseStream.Position += 4;
                    map.Index.Read(map);
                    map.TagHandler.Read(map);
                    map.SelectedTag = map.FindTag(map.SelectedTag.Identifier);
                }
            }

            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 68;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            map.bw.BaseStream.Position = r.Translation + 4;
            map.bw.Write(Chunk.Width);
            map.bw.Write(Chunk.Height);
            map.bw.Write(Chunk.Depth);
            map.bw.BaseStream.Position += 4;
            map.bw.Write((int)Chunk.BitmapFormat);
            map.bw.BaseStream.Position = r.Translation + 28;
            map.bw.Write(Chunk.LOD1);
            map.bw.BaseStream.Position = r.Translation + 48;
            map.bw.Write(Chunk.SourceData.Length);
        }

        public static void Internalize(Map map, BitmapChunk Chunk)
        {
            map.br.BaseStream.Position = map.Header.IndexOffset;
            byte[] EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
            map.bw.BaseStream.Position = map.Header.IndexOffset;
            int Offset = (int)map.bw.BaseStream.Position;
            map.bw.Write(Chunk.SourceData);
            int Padding = map.Padding();
            map.bw.Write(new byte[Padding]);
            map.bw.Write(EOF);
            map.Header.IndexOffset += Chunk.SourceData.Length + Padding;
            map.bw.BaseStream.Position = 8;
            map.bw.Write((int)map.bw.BaseStream.Length);
            map.bw.BaseStream.Position += 4;
            map.bw.Write(map.Header.IndexOffset);
            map.Index.Read(map);
            map.TagHandler.Read(map);
            map.SelectedTag = map.FindTag(map.SelectedTag.Identifier);
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 68;
            Reflexive r = new Reflexive();
            r.Read(map, true);
            map.bw.BaseStream.Position = r.Translation + 28;
            map.bw.Write(Offset);
            map.bw.Write((int)-1);
            map.bw.Write((int)-1);
            map.bw.BaseStream.Position = r.Translation + 48;
            map.bw.Write(Chunk.SourceData.Length);
            map.bw.Write(0);
            map.bw.Write(0);
        }
        #endregion
    }
}
