﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using HaloMap.Models;

namespace HaloMap.Models
{
    #region Model Raw Layout
    public class ModelResourceBlock
    {
        #region Fields
        public short ChunkSize, ChunkCount;
        public int Offset, Size;
        #endregion

        #region Methods
        public void Read(BinaryReader br)
        {
            br.BaseStream.Position += 4;
            ChunkCount = br.ReadInt16();
            ChunkSize = br.ReadInt16();
            Size = br.ReadInt32();
            Offset = br.ReadInt32();
        }
        #endregion
    }

    public class MeshInfoBlock : ModelResourceBlock
    {
        #region Fields
        public int Unknown;
        public List<short> ShaderIndex = new List<short>();
        public List<short> IndiceStart = new List<short>();
        public List<short> IndiceCount = new List<short>();
        #endregion

        #region Constructor
        public MeshInfoBlock(BinaryReader br)
        {
            base.Read(br);
        }
        #endregion

        #region Methods
        public void Read(BinaryReader br, int Offset)
        {
            br.BaseStream.Position = Offset + base.Offset;
            Unknown = br.ReadInt32();
            for (int i = 0; i < base.ChunkCount; i++)
            {
                ShaderIndex.Add(br.ReadInt16());
                IndiceStart.Add(br.ReadInt16());
                IndiceCount.Add(br.ReadInt16());
                br.BaseStream.Position += 66;
            }
        }
        #endregion
    }

    public class IndiceBlock : ModelResourceBlock
    {
        #region Fields
        public List<short> Indice = new List<short>();
        #endregion

        #region Constructor
        public IndiceBlock(BinaryReader br)
        {
            base.Read(br);
        }
        #endregion

        #region Methods
        public void Read(BinaryReader br, int Offset)
        {
            br.BaseStream.Position = Offset + base.Offset;
            for (int i = 0; i < base.ChunkCount; i++)
            {
                Indice.Add(br.ReadInt16());
            }
        }
        #endregion
    }

    public class UnknownBlock : ModelResourceBlock
    {
        #region Fields
        public List<int> Unknown = new List<int>();
        #endregion

        #region Constructor
        public UnknownBlock(BinaryReader br)
        {
            base.Read(br);
        }
        #endregion

        #region Methods
        public void Read(BinaryReader br, int Offset)
        {
            br.BaseStream.Position = Offset + base.Offset;
            for (int i = 0; i < base.ChunkCount; i++)
            {
                Unknown.Add(br.ReadInt32());
                br.BaseStream.Position += 28;
            }
        }
        #endregion
    }

    public class BoneInfo
    {
        #region Fields
        public List<int> BoneIndex = new List<int>();
        public List<float> Weight = new List<float>();
        #endregion
    }

    public class VerticeBlock : ModelResourceBlock
    {
        #region Fields
        public List<Vector3> Verts = new List<Vector3>();
        public List<BoneInfo> Bones = new List<BoneInfo>();
        #endregion

        #region Constructor
        public VerticeBlock(BinaryReader br)
        {
            base.Read(br);
        }
        #endregion

        #region Methods
        public void Read(BinaryReader br, BoundingBox bb, int Offset)
        {
            br.BaseStream.Position = Offset + base.Offset;
            for (int i = 0; i < base.ChunkCount; i++)
            {
                Vector3 Vert = new Vector3();
                Vert.X = Model.DecompressVertice(br.ReadSingle(), bb.XMin, bb.XMax);
                Vert.Y = Model.DecompressVertice(br.ReadSingle(), bb.YMin, bb.YMax);
                Vert.Z = Model.DecompressVertice(br.ReadSingle(), bb.ZMin, bb.ZMax);
                Verts.Add(Vert);
                switch (base.ChunkSize)
                {
                    case 6:
                        {
                            BoneInfo b = new BoneInfo();
                            b.BoneIndex.Add(0);
                            b.Weight.Add(1.0f);
                            Bones.Add(b);
                            break;
                        }
                    case 8:
                        {
                            BoneInfo c = new BoneInfo();
                            c.BoneIndex.Add(br.ReadByte());
                            c.Weight.Add(1.0f);
                            byte tempb = br.ReadByte();
                            if (tempb == 0)
                            {
                                Bones.Add(c);
                                break;
                            }
                            BoneInfo c2 = new BoneInfo();
                            c2.BoneIndex.Add(tempb);
                            c2.Weight.Add(1.0f);
                            Bones.Add(c2);
                            c.Weight[0] = 1.0f;
                            Bones.Add(c);
                            break;
                        }
                    case 12:
                        {
                            BoneInfo bbb = new BoneInfo();
                            bbb.BoneIndex.Add(br.ReadByte());
                            bbb.Weight.Add(0.99f);
                            Bones.Add(bbb);
                            break;
                        }
                }
            }
        }
        #endregion
    }

    public class UVData : ModelResourceBlock
    {
        #region Fields
        public List<Vector2> UVs = new List<Vector2>();
        #endregion

        #region Constructor
        public UVData(BinaryReader br)
        {
            base.Read(br);
        }
        #endregion

        #region Methods
        public void Read(BinaryReader br, BoundingBox bb, int Offset)
        {
            br.BaseStream.Position = Offset + base.Offset;
            for (int i = 0; i < base.ChunkCount; i++)
            {
                Vector2 uv = new Vector2();
                uv.X = Model.DecompressVertice(br.ReadSingle(), bb.XMin, bb.XMax) % 1;
                if (uv.X < 0) { uv.X = 1 - uv.X; }
                uv.Y = Model.DecompressVertice(br.ReadSingle(), bb.YMin, bb.YMax) % 1;
                if (uv.Y < 0) { uv.Y = 1 - uv.Y; }
                UVs.Add(uv);
            }
        }
        #endregion
    }

    public class NormalBlock : ModelResourceBlock
    {
        #region Fields
        public List<Vector3> Normals = new List<Vector3>();
        #endregion

        #region Constructor
        public NormalBlock(BinaryReader br)
        {
            base.Read(br);
        }
        #endregion

        #region Methods
        public void Read(BinaryReader br, int Offset)
        {
            br.BaseStream.Position = Offset + base.Offset;
            for (int i = 0; i < base.ChunkCount; i++)
            {
                br.BaseStream.Position = Offset + base.Offset + (i * base.ChunkSize);
                Vector3 Normal = Model.DecompressNormal(br.ReadInt32());
                Normals.Add(Normal);
            }
        }
        #endregion
    }

    public class BoneMap : ModelResourceBlock
    {
        #region Fields
        public List<byte> BoneMaps = new List<byte>();
        #endregion

        #region Constructor
        public BoneMap(BinaryReader br)
        {
            base.Read(br);
        }
        #endregion

        #region Methods
        public void Read(BinaryReader br, int Offset)
        {
            br.BaseStream.Position = Offset + base.Offset;
            for (int i = 0; i < base.ChunkCount; i++)
            {
                BoneMaps.Add(br.ReadByte());
            }
        }
        #endregion
    }

    public class ModelRawBlock
    {
        #region Fields
        public int Offset, Size;
        public string BlockHeader;
        public int DataSize, IndiceCount, UnknownBlockCount, BoneMapCount;
        public RawLocation Location;
        public MeshInfoBlock MeshInfo;
        public IndiceBlock Indices;
        public UnknownBlock UnknownBlock;
        public VerticeBlock Vertices;
        public UVData UVData;
        public NormalBlock Normals;
        public BoneMap BoneMaps;
        public BoundingBox bb;
        #endregion

        #region Constructor
        public ModelRawBlock(int Offset, int Size, RawLocation Location, BoundingBox bb)
        {
            this.Offset = Offset;
            this.Size = Size;
            this.Location = Location;
            this.bb = bb;
        }
        #endregion

        #region Methods
        public void Read(Map map)
        {
            map.br.BaseStream.Position += 16;
            MeshInfo = new MeshInfoBlock(map.br);
            Indices = new IndiceBlock(map.br);
            UnknownBlock = new UnknownBlock(map.br);
            Vertices = new VerticeBlock(map.br);
            UVData = new UVData(map.br);
            Normals = new NormalBlock(map.br);
            BoneMaps = new BoneMap(map.br);

            BinaryReader br = map.OpenMap(Location);
            MeshInfo.Read(br, Offset + 120);
            Indices.Read(br, Offset + 120);
            UnknownBlock.Read(br, Offset + 120);
            Vertices.Read(br, bb, Offset + 120);
            UVData.Read(br, bb, Offset + 120);
            Normals.Read(br, Offset + 120);
            BoneMaps.Read(br, Offset + 120);
        }
        #endregion
    }
    #endregion
}

namespace HaloMap
{
    public class ModelMetaBlock
    {
        #region Fields
        public short VertexCount, FaceCount;
        public int RawOffset, RawSize, HeaderSize, DataSize;
        public RawLocation Location;
        public ModelRawBlock RawData;
        #endregion

        #region Methods
        public void Read(Map map, BoundingBox bb)
        {
            map.br.BaseStream.Position += 4;
            VertexCount = map.br.ReadInt16();
            FaceCount = map.br.ReadInt16();
            map.br.BaseStream.Position += 48;
            RawOffset = map.br.ReadInt32();
            RawSize = map.br.ReadInt32();
            HeaderSize = map.br.ReadInt32();
            DataSize = map.br.ReadInt32();
            Location = HaloMap.RawData.RawLocator(ref RawOffset);
            Reflexive Resources = new Reflexive();
            Resources.Read(map, true);
            map.br.BaseStream.Position = Resources.Translation;
            RawData = new ModelRawBlock(RawOffset, RawSize, Location, bb);
            RawData.Read(map);
        }
        #endregion
    }

    public class BoundingBox
    {
        #region Fields
        public float XMin;
        public float XMax;
        public float YMin;
        public float YMax;
        public float ZMin;
        public float ZMax;
        public float UMin;
        public float UMax;
        public float VMin;
        public float VMax;
        #endregion

        #region Methods
        public void Read(Map map)
        {
            XMin = map.br.ReadSingle();
            XMax = map.br.ReadSingle();
            YMin = map.br.ReadSingle();
            YMax = map.br.ReadSingle();
            XMin = map.br.ReadSingle();
            XMax = map.br.ReadSingle();
            ZMin = map.br.ReadSingle();
            ZMax = map.br.ReadSingle();
            UMin = map.br.ReadSingle();
            UMax = map.br.ReadSingle();
            VMin = map.br.ReadSingle();
            VMax = map.br.ReadSingle();
        }
        #endregion
    }

    public class Model
    {
        #region Fields
        public BoundingBox BoundingBox = new BoundingBox();
        public List<ModelMetaBlock> Model1 = new List<ModelMetaBlock>();
        #endregion

        #region Constructor
        public Model(Map map)
        {
            //Bounding Box
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 20;
            Reflexive BB = new Reflexive();
            BB.Read(map, true);
            map.br.BaseStream.Position = BB.Translation;
            BoundingBox.Read(map);

            //Raw Data
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 36;
            Reflexive Sections = new Reflexive();
            Sections.Read(map, true);
            for (int i = 0; i < Sections.ChunkCount; i++)
            {
                map.br.BaseStream.Position = Sections.Translation + (i * 92);
                ModelMetaBlock block = new ModelMetaBlock();
                block.Read(map, BoundingBox);
                Model1.Add(block);
            }
        }
        #endregion

        #region Methods
        public void DrawMeshes(ref Device device)
        {
            int VCount = 0;
            for (int i = 0; i < Model1.Count; i++)
                VCount += Model1[i].RawData.Vertices.Verts.Count;
            Mesh[] meshes = new Mesh[Model1.Count];
            VertexBuffer vb = new VertexBuffer(typeof(HaloVertex), VCount, device, Usage.WriteOnly, HaloVertex.FVF, Pool.Managed);
            HaloVertex[] verts = (HaloVertex[])vb.Lock(0, 0); // Lock the buffer (which will return our structs)
            device.VertexFormat = HaloVertex.FVF;
            for (int x = 0; x < Model1.Count; x++)
            {
                for (int i = 0; i < Model1[x].VertexCount; i++)
                {
                    verts[i].Position = new Vector3(Model1[x].RawData.Vertices.Verts[i].X, Model1[x].RawData.Vertices.Verts[i].Y, Model1[x].RawData.Vertices.Verts[i].Z);
                    verts[i].Tu0 = Model1[x].RawData.UVData.UVs[i].X;
                    verts[i].Tv0 = Model1[x].RawData.UVData.UVs[i].Y;
                    verts[i].Normal = new Vector3(Model1[x].RawData.Normals.Normals[i].X, Model1[x].RawData.Normals.Normals[i].Y, Model1[x].RawData.Normals.Normals[i].Z);
                    verts[i].specular = 1;
                    verts[i].diffuse = 1;

                    ModelMetaBlock mmb = Model1[x];
                    meshes[i] = Rendering.MakeMesh(ref mmb, ref device);
                }
            }
            vb.Unlock();

            device.RenderState.FillMode = FillMode.Solid;
            device.SetStreamSource(0, vb, 0);
            for (int i = 0; i < Model1.Count; i++)
            {
                meshes[i].DrawSubset(i);
            }
        }

        public static float DecompressVertice(float input, float min, float max)
        {
            float percent = (input + 32768) / 65535;
            float result = (percent * (max - min)) + min;
            return result;
        }

        public static float CompressVertice(float input, float min, float max)
        {
            float result = input - min;
            result = (result / (max - min));
            result = result * 65535;
            result = result - 32768;
            return result;
        }

        public static Vector3 DecompressNormal(int compressednormal)
        {
            int xx = compressednormal & 0x3ff;
            float xxx = Convert.ToSingle(xx) / Convert.ToSingle(0x3ff);
            int bitx = ((compressednormal >> 10) & 0x1);
            if (bitx == 1) { xxx = 1 - xxx; xxx = -xxx; }


            int yy = ((compressednormal >> 11) & 0x3ff);
            float yyy = Convert.ToSingle(yy) / Convert.ToSingle(0x3ff);
            int bity = ((compressednormal >> 21) & 0x1);
            if (bity == 1) { yyy = 1 - yyy; yyy = -yyy; }


            int zz = ((compressednormal >> 22) & 0x1ff);
            float zzz = Convert.ToSingle(zz) / Convert.ToSingle(0x1ff);
            int bitz = ((compressednormal >> 31) & 0x1);
            if (bitz == 1) { zzz = 1 - zzz; zzz = -zzz; }
            return new Vector3(xxx, yyy, zzz);
        }

        public static int CompressNormal(Vector3 normal)
        {
            bool negx = false;
            if (normal.X < 0) { negx = true; normal.X = -normal.X; }
            float bbb = normal.X * Convert.ToSingle(0x3ff);
            int b = Convert.ToInt32(bbb);
            if (negx) b = -b + 1 | 0x400;

            bool negy = false;
            if (normal.Y < 0) { negy = true; normal.Y = -normal.Y; }
            float bbb2 = normal.Y * Convert.ToSingle(0x3ff);
            int b2 = Convert.ToInt32(bbb2);
            if (negy) b2 = -b2 + 1 | 0x400;
            b2 <<= 11;


            bool negz = false;
            if (normal.Z < 0) { negz = true; normal.Z = -normal.Z; }
            float bbb3 = normal.Z * Convert.ToSingle(0x1ff);
            int b3 = Convert.ToInt32(bbb3);
            if (negz) b3 = -b3 + 1 | 0x200;
            b3 <<= 22;


            int result = b3 | b2 | b;

            return result;
        }
        #endregion
    }
}
